/*      
** avl_tree.c	- AVL Tree Library
**
** Copyright (c) 1996  Hughes Technologies Pty Ltd
**
*************************************************************************
**
** This library implements a balanced tree capable of holding key/offset
** pairs.  The offset is assumed to be used as an offset into a data table.
** The library supports user defined key lengths on a per tree basis, and
** support for multiplte data types (i.e. int's, char strings and reals are
** inserted into the tree in the correct order, not just the order of the
** bytes that comprise the value).  This allows the btree to be used for
** range indexing.  Duplicate keys are supported.
**
** This library was written for Mini SQL.
**
** Mods for mSQL 2.1.x - Data stats are mainteined in the super block 
** including the total number of entries and the number of unique keys.
** The engine can use that info to determine index density when working
** on "union index" creation.
**
*/


static int 	AVL_DEBUG = 0,
		TEST_TREE = 0;

#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdint.h>
#include <sys/types.h>
#include <sys/mman.h>
#include <sys/stat.h>

#include <common/config.h>
#include <common/config_extras.h>
#include <common/portability.h>

#ifdef HAVE_STRINGS_H
#  include <strings.h>
#endif
#ifdef HAVE_STDINT_H
#  include <stdint.h>
#endif


#include <common/debug/debug.h>

#include "index.h"
#include "avl_tree.h"

#define	AVL_MAGIC_V1 		0x01020304
#define	AVL_MAGIC_V2 		0x02020304
#define	AVL_INVALID 		0xAABBCCDD

#define	SBLK_SIZE		(sizeof(avl_sbk) + (8 - (sizeof(avl_sbk)% 8)))
#define	PSBLK_SIZE		(sizeof(ptr_sbk) + (8 - (sizeof(ptr_sbk)% 8)))
#define DISK_NODE_SIZE(L)	(sizeof(avl_nod)+(L))
#define OFF_TO_NODE(t,o)	((o - SBLK_SIZE) / t->sblk->nodeLen + 1)
#define NODE_TO_OFF(t,n)	(((n - 1) * t->sblk->nodeLen) + SBLK_SIZE)
#define	REG			register


#define bcopy4(s,d) \
      ((((unsigned char *)d)[3] = ((unsigned char *)s)[3]), \
       (((unsigned char *)d)[2] = ((unsigned char *)s)[2]), \
       (((unsigned char *)d)[1] = ((unsigned char *)s)[1]), \
       (((unsigned char *)d)[0] = ((unsigned char *)s)[0]))

#define bcopy8(s,d) \
      ((((unsigned char *)d)[7] = ((unsigned char *)s)[7]), \
       (((unsigned char *)d)[6] = ((unsigned char *)s)[6]), \
       (((unsigned char *)d)[5] = ((unsigned char *)s)[5]), \
       (((unsigned char *)d)[4] = ((unsigned char *)s)[4]), \
       (((unsigned char *)d)[3] = ((unsigned char *)s)[3]), \
       (((unsigned char *)d)[2] = ((unsigned char *)s)[2]), \
       (((unsigned char *)d)[1] = ((unsigned char *)s)[1]), \
       (((unsigned char *)d)[0] = ((unsigned char *)s)[0]))


#define nodeKey(node) \
	(char*)(((char*)node) + sizeof(avl_nod))


static void avlDebugTrap()
{
}


/*************************************************************************
**************************************************************************
**                                                                      **
**                         UTILITY ROUTINES                             **
**                                                                      **
**************************************************************************
*************************************************************************/

#define DISK_EXTEND_LENGTH		(1024 * 5)
#define PTR_EXTEND_LENGTH		(1024 * 5)
/*
#define DISK_EXTEND_LENGTH		1024
#define PTR_EXTEND_LENGTH		1024
*/
#define MEM_EXTEND_LENGTH		256

static int extendTree(tree)
	avltree	*tree;
{
	int	nodeLen,
		count,
		extendLen;
	off_t	nodeOff,
		curOff;
	avl_nod	*cur;


	if (AVL_DEBUG > 0)
	{
		printf("Extending tree\n");
	}
	if (tree->memTree == NULL)
	{
		/*
		** Unmap the file, extend it and then re-map it
		*/
		extendLen = DISK_EXTEND_LENGTH;
		tree->remapped = 1;
		nodeLen = tree->sblk->nodeLen;
		munmap(tree->treeMap, tree->treeMapLen);
		lseek(tree->treeFd, (extendLen * nodeLen) - 1, SEEK_END );
		if (write(tree->treeFd, "\0", 1) < 0)
		{
			perror("write failed");
			return(-1);
		}
		nodeOff = tree->treeMapLen;
		tree->treeMapLen += extendLen * nodeLen;
		tree->treeMap = mmap(NULL, tree->treeMapLen, 
			PROT_READ|PROT_WRITE, MAP_SHARED, tree->treeFd, 
			(off_t)0);
		if (tree->treeMap == (void*)-1)
		{
			char msg[160];
			sprintf(msg,"mmap of %u failed",(unsigned int)
				tree->treeMapLen);
			perror(msg);
			return(-1);
		}
		tree->sblk = (avl_sbk *)tree->treeMap;
		tree->sblk->fileLen = tree->treeMapLen;
	}
	else
	{
		extendLen = MEM_EXTEND_LENGTH;
		nodeOff = tree->memLen;
		nodeLen = tree->sblk->nodeLen;
		tree->memTree = realloc(tree->memTree,tree->memLen +
			(extendLen * nodeLen));
		if (tree->memTree == (void*)-1)
		{
			char msg[160];
			sprintf(msg,"mmap of %u failed",(unsigned int)
				tree->treeMapLen);
			perror(msg);
			return(-1);
		}
		bzero(tree->memTree + tree->memLen, extendLen * nodeLen);
		tree->sblk = (avl_sbk *)tree->memTree;
		tree->treeMap = (char *)tree->memTree;
		tree->memLen += extendLen * tree->sblk->nodeLen;
	}
	
	/*
	** Setup the free list to wind through the new nodes
	*/
	curOff = nodeOff;
	for (count = 0; count < extendLen; count++)
	{
		cur = (avl_nod *)(tree->treeMap + curOff);
		cur->parent = tree->sblk->freeList;
		tree->sblk->freeList = OFF_TO_NODE(tree, curOff);
		curOff += tree->sblk->nodeLen;
	}
	return(0);
}



static int checkTreeRemap(tree)
	avltree	*tree;
{
	off_t	oldLen,
		newLen;

	if (tree->sblk->fileLen > tree->treeMapLen)
	{
		oldLen = tree->treeMapLen;
		newLen = tree->sblk->fileLen;
		munmap(tree->treeMap, oldLen);
		tree->treeMap = mmap(NULL, newLen, PROT_READ|PROT_WRITE,
			MAP_SHARED, tree->treeFd, (off_t)0);
		if (tree->treeMap == (caddr_t)-1)
		{
			perror("tree mmap");
			return(-1);
		}
		tree->sblk = (avl_sbk *)tree->treeMap;
		tree->treeMapLen = newLen;
	}

	if (tree->version == 1)
	{
		return(0);
	}

	if (tree->psblk->fileLen > tree->ptrMapLen)
	{
		oldLen = tree->ptrMapLen;
		newLen = tree->psblk->fileLen;
		munmap(tree->ptrMap, oldLen);
		tree->ptrMap = mmap(NULL, newLen, PROT_READ|PROT_WRITE, 
			MAP_SHARED, tree->ptrFd, (off_t)0);
		if (tree->ptrMap == (caddr_t)-1)
		{
			perror("ptr mmap");
			return(-1);
		}
		tree->psblk = (ptr_sbk *)tree->ptrMap;
		tree->ptrMapLen = newLen;
	}
	return(0);
}

	

static avl_nod *getFreeNode(tree)
	avltree	*tree;
{
	int	freeNode = 0;
	avl_nod	*cur;


	/*
	** Moved to the start of each API call so we don't remap
	** in the middle of a tree traversal
	if (checkTreeRemap(tree) < 0)
	{
		return(NULL);
	}
	*/


	if (tree->sblk->freeList == 0)
	{
		if (extendTree(tree) < 0)
			return(NULL);
	}
	freeNode = tree->sblk->freeList;
	cur = (avl_nod*)(tree->treeMap + NODE_TO_OFF(tree,freeNode));
	tree->sblk->freeList = cur->parent;
	cur->left = cur->right = cur->parent = cur->height = cur->dup = 0;
	/*cur->key = (char *)((char *)cur)+sizeof(avl_nod);*/
	cur->nodeNum = freeNode;
	return(cur);
}



static void dropNode(tree,cur)
        avltree	*tree;
        avl_nod	*cur;
{
        cur->parent = tree->sblk->freeList;
        tree->sblk->freeList = cur->nodeNum;
}



static avl_nod *getTreeNode(tree, num)
	avltree	*tree;
	uint	num;
{
	avl_nod	*cur;
	char	testChar;

	if (num == 0)
		return(NULL);

	/*
	if (checkTreeRemap(tree) < 0)
	{
		return(NULL);
	}
	*/
	cur = (avl_nod *)(tree->treeMap + NODE_TO_OFF(tree, num));
	/*cur->key = ((char *)cur) + sizeof(avl_nod);*/
	return(cur);
}



void avlPrintElement(data,type,len)
	char	*data;
	int	type,
		len;
{
	switch(type)
	{
		case IDX_CHAR:
			printf("char '%s'",data);
			break;

		case IDX_BYTE:
			{
				int	count = 0;
				printf("byte ");
				while(count < len)
				{
					printf("%u.", (u_char)*(data+count));
					count++;
				}
			}
			break;

		case IDX_INT8:
			printf("int8 '%hhd'",(int8_t)*(int8_t*)data);
			break;

		case IDX_INT16:
			printf("int16 '%hd'",(int16_t)*(int16_t*)data);
			break;

		case IDX_INT32:
			printf("int '%d'",(int)*(int*)data);
			break;

		case IDX_INT64:
			printf("int64 '%lld'",(int64_t)*(int64_t*)data);
			break;
			
		case IDX_UINT8:
			printf("uint8 '%hhu'",(uint8_t)*(uint8_t*)data);
			break;

		case IDX_UINT16:
			printf("uint16 '%hu'",(uint16_t)*(uint16_t*)data);
			break;

		case IDX_UINT32:
			printf("uint '%u'",(u_int)*(u_int*)data);
			break;

		case IDX_UINT64:
			printf("uint64 '%llu'",(uint64_t)*(uint64_t*)data);
			break;

		case IDX_REAL:
			printf("real '%f'",(double)*(double*)data);
			break;

		default:
			printf("Unknown Type!!!");
			break;
	}
}


static void copyValue(src,dst,tree)
	char	*src,
		*dst;
	avltree	*tree;
{
	switch(tree->sblk->keyType)
	{
		case IDX_INT8:
		case IDX_UINT8:
			bcopy(src,dst,1);
			break;
		case IDX_INT16:
		case IDX_UINT16:
			bcopy(src,dst,2);
			break;
		case IDX_INT32:
		case IDX_UINT32:
			bcopy4(src,dst);
			break;
		case IDX_REAL:
			bcopy8(src,dst);
			break;
		case IDX_CHAR:
			strncpy(dst,src,tree->sblk->keyLen-1);
			*(dst+tree->sblk->keyLen-1)=0;
			break;
		default:
			bcopy(src, dst, tree->sblk->keyLen);
	}
}




static void showIndent(d)
	int	d;
{
	int	i;

	for(i = 0; i < d; i++)
	{
		printf("   ");
	}
}



static int getNodeHeight(node)
	avl_nod	*node;
{
	return( (node)? node->height : -1);
}



#define setNodeHeight(tree, node)			\
{							\
	int	_leftH,					\
		_rightH;				\
	avl_nod	*_lNode,				\
		*_rNode;				\
							\
	if (node)					\
	{						\
		_lNode = getTreeNode(tree,node->left);	\
		_leftH = getNodeHeight(_lNode);		\
		_rNode = getTreeNode(tree,node->right);	\
		_rightH = getNodeHeight(_rNode);	\
		if (_leftH > _rightH)			\
			node->height = 1 + _leftH;	\
		else					\
			node->height = 1 + _rightH;	\
	}						\
}


static void rotateRight(tree, node)
	avltree	*tree;
	avl_nod	*node;
{
	avl_nod	*parent,
		*left,
		*tmp;

	if (AVL_DEBUG > 1)
	{
		printf("Right rotate @ ");
		avlPrintElement(nodeKey(node),tree->sblk->keyType,
			tree->sblk->keyLen);
		printf("\n");
	}

	parent = getTreeNode(tree,node->parent);
	left = getTreeNode(tree, node->left);
	node->left = left->right;
	tmp = getTreeNode(tree,node->left);
	if (tmp)
		tmp->parent = node->nodeNum;
	left->right = node->nodeNum;
	if (parent)
	{
		if (parent->left == node->nodeNum)
			parent->left = left->nodeNum;
		else
			parent->right = left->nodeNum;
	}
	else
	{
		tree->sblk->root = left->nodeNum;
	}
	left->parent = node->parent;
	node->parent = left->nodeNum;
	setNodeHeight(tree,left);
	setNodeHeight(tree,node);
}



static void rotateLeft(tree, node)
	avltree	*tree;
	avl_nod	*node;
{
	avl_nod	*parent,
		*right,
		*tmp;

	if (AVL_DEBUG > 1)
	{
		printf("Left rotate @ ");
		avlPrintElement(nodeKey(node),tree->sblk->keyType,
			tree->sblk->keyLen);
		printf("\n");
	}
	parent = getTreeNode(tree,node->parent);
	right = getTreeNode(tree, node->right);
	node->right = right->left;
	tmp = getTreeNode(tree,node->right);
	if (tmp)
		tmp->parent = node->nodeNum;
	right->left = node->nodeNum;
	if (parent)
	{
		if (parent->right == node->nodeNum)
			parent->right = right->nodeNum;
		else
			parent->left = right->nodeNum;
	}
	else
	{
		tree->sblk->root = right->nodeNum;
	}
	right->parent = node->parent;
	node->parent = right->nodeNum;
	setNodeHeight(tree,right);
	setNodeHeight(tree,node);
}



static void doubleRight(tree, node)
	avltree	*tree;
	avl_nod	*node;
{
	avl_nod	*tmp;

	tmp = getTreeNode(tree, node->left);
	rotateLeft(tree,tmp);
	rotateRight(tree,node);
}


static void doubleLeft(tree, node)
	avltree	*tree;
	avl_nod	*node;
{
	avl_nod	*tmp;

	tmp = getTreeNode(tree,node->right);
	rotateRight(tree,tmp);
	rotateLeft(tree,node);
}

static void balanceTree(tree, node)
	avltree	*tree;
	avl_nod	*node;
{
	int	leftHeight,
		rightHeight;
	avl_nod	*left,
		*right,
		*child;

	left = getTreeNode(tree, node->left);
	right = getTreeNode(tree, node->right);
	leftHeight = getNodeHeight(left);
	rightHeight = getNodeHeight(right);

	if (leftHeight > rightHeight + 1)
	{
		child = getTreeNode(tree, node->left);
		left = getTreeNode(tree,child->left);
		right = getTreeNode(tree,child->right);
		leftHeight = getNodeHeight(left);
		rightHeight = getNodeHeight(right);

		if (leftHeight >= rightHeight)
		{
			rotateRight(tree,node);
		}
		else
		{
			doubleRight(tree,node);
		}
	}
	else if (rightHeight > leftHeight + 1)
	{
		child = getTreeNode(tree, node->right);
		left = getTreeNode(tree,child->left);
		right = getTreeNode(tree,child->right);
		leftHeight = getNodeHeight(left);
		rightHeight = getNodeHeight(right);

		if (rightHeight >= leftHeight)
			rotateLeft(tree,node);
		else
			doubleLeft(tree,node);
	}
}



static int extendPtr(tree)
	avltree	*tree;
{
	off_t	oldMapLen;
	char	*cp;

	/*
	** Extend and remap the ptr file
	*/
	munmap(tree->ptrMap, tree->ptrMapLen);
	lseek(tree->ptrFd, (PTR_EXTEND_LENGTH * sizeof(int)) - 1, SEEK_END );
	if (write(tree->ptrFd, "\0", 1) < 0)
	{
		perror("write failed");
		return(-1);
	}
	oldMapLen = tree->ptrMapLen;
	tree->ptrMapLen += (PTR_EXTEND_LENGTH * sizeof(int));
	tree->ptrMap = mmap(NULL, tree->ptrMapLen, PROT_READ|PROT_WRITE, 
		MAP_SHARED, tree->ptrFd, (off_t)0);
	if (tree->ptrMap == (void*)-1)
	{
		char msg[160];
		sprintf(msg,"mmap of ptr %u failed",
			(unsigned int) tree->ptrMapLen);
		perror(msg);
		return(-1);
	}
	tree->psblk = (ptr_sbk *)tree->ptrMap;

	/*
	** Initialise the new entries to #FFFFFFFF
	*/
	cp = tree->ptrMap + oldMapLen;
	memset(cp, 255, tree->ptrMapLen - oldMapLen);
	return(0);
}


static int addPtrEntry(tree, data, nodeNum)
	avltree	*tree;
	off_t	data;
	int	nodeNum;
{
	off_t	offset,
		oldMapLen;
	char	*cp;

	if (tree->version < 2)
		return(-1);


	/*
	** Find the place to write this pointer entry.  Extend the 
	** ptr file if needed
	*/
	offset = (data * sizeof(int)) + PSBLK_SIZE;

	while (offset >= tree->ptrMapLen)
	{
		if (extendPtr(tree) < 0)
			return(-1);
	}

	cp = tree->ptrMap + offset;
	bcopy4(&nodeNum,cp);
	return(0);
}



static int delPtrEntry(tree, data)
	avltree	*tree;
	off_t	data;
{
	off_t	offset,
	oldMapLen;
	char	*cp;
	int	nodeNum;

	if (tree->version < 2)
		return(-1);

	offset = (data * sizeof(u_int)) + PSBLK_SIZE;
	if (offset >= tree->ptrMapLen)
		return(-1);
	cp = tree->ptrMap + offset;
	nodeNum = -1;
	bcopy4(&nodeNum,cp);
	return(0);
}

static int getPtrEntry(tree, data)
	avltree	*tree;
	off_t	data;
{
	int	nodeNum;
	off_t	offset;
	char	*cp;

	offset = (data * sizeof(int)) + PSBLK_SIZE;
	cp = (char*) tree->ptrMap + offset;
	bcopy4(cp, &nodeNum);
	return(nodeNum);
}



/*************************************************************************
**************************************************************************
**                                                                      **
**                       TYPE HANDLING ROUTINES                         **
**                                                                      **
**************************************************************************
*************************************************************************/


static int int8Compare(v1,v2)
	char	*v1,
		*v2;
{
	int8_t	int1,
		int2;

	bcopy(v1,&int1,1);
	bcopy(v2,&int2,1);
	if (int1 > int2)
		return(1);
	if (int1 < int2)
		return(-1);
	return(0);
}


static int uint8Compare(v1,v2)
	char	*v1,
		*v2;
{
	uint8_t	int1,
		int2;

	bcopy(v1,&int1,1);
	bcopy(v2,&int2,1);
	if (int1 > int2)
		return(1);
	if (int1 < int2)
		return(-1);
	return(0);
}

static int int16Compare(v1,v2)
	char	*v1,
		*v2;
{
	int16_t	int1,
		int2;

	bcopy(v1,&int1,2);
	bcopy(v2,&int2,2);
	if (int1 > int2)
		return(1);
	if (int1 < int2)
		return(-1);
	return(0);
}


static int uint16Compare(v1,v2)
	char	*v1,
		*v2;
{
	uint16_t int1,
		int2;

	bcopy(v1,&int1,2);
	bcopy(v2,&int2,2);
	if (int1 > int2)
		return(1);
	if (int1 < int2)
		return(-1);
	return(0);
}

static int int32Compare(v1,v2)
	char	*v1,
		*v2;
{
	int	int1,
		int2;

	bcopy4(v1,&int1);
	bcopy4(v2,&int2);
	if (int1 > int2)
		return(1);
	if (int1 < int2)
		return(-1);
	return(0);
}


static int uint32Compare(v1,v2)
	char	*v1,
		*v2;
{
	u_int	int1,
		int2;

	bcopy4(v1,&int1);
	bcopy4(v2,&int2);
	if (int1 > int2)
		return(1);
	if (int1 < int2)
		return(-1);
	return(0);
}

static int int64Compare(v1,v2)
	char	*v1,
		*v2;
{
	int64_t	int1,
		int2;

	bcopy8(v1,&int1);
	bcopy8(v2,&int2);
	if (int1 > int2)
		return(1);
	if (int1 < int2)
		return(-1);
	return(0);
}


static int uint64Compare(v1,v2)
	char	*v1,
		*v2;
{
	uint64_t int1,
		int2;

	bcopy8(v1,&int1);
	bcopy8(v2,&int2);
	if (int1 > int2)
		return(1);
	if (int1 < int2)
		return(-1);
	return(0);
}



static int realCompare(v1,v2)
	char	*v1,
		*v2;
{
	double	r1,
		r2;

	bcopy8(v1,&r1);
	bcopy8(v2,&r2);
	if (r1 > r2)
		return(1);
	if (r1 < r2)
		return(-1);
	return(0);
}



static int charCompare(v1,v2)
	char	*v1,
		*v2;
{
	register char	*cp1,
		*cp2;

	cp1 = v1;
	cp2 = v2;
	while (*cp1 == *cp2)
	{
		if (*cp1 == 0)
		{
			return(0);
		}
		cp1++;
		cp2++;
	}
	if ((unsigned char )*cp1 > (unsigned char)*cp2)
		return(1);
	return(-1);
}



#ifdef NOTDEF

/* Use local memcmp() instead as it'll be optimised for the hardware */
static int byteCompare(v1,v2,len)
	char	*v1,
		*v2;
	int	len;
{
	char	*cp1,
		*cp2;

	cp1 = v1;
	cp2 = v2;
	while (len && *cp1 == *cp2)
	{
		cp1++;
		cp2++;
		len--;
	}
	if (len == 0)
		return(0);
	if ( (unsigned char) *cp1 > (unsigned char)*cp2)
		return(1);
	return(-1);
}
#endif



static int compareValues(v1,v2,tree)
	char	*v1,
		*v2;
	avltree	*tree;
{
	int	res;

	switch(tree->sblk->keyType)
	{
		case IDX_BYTE:
			/*return(byteCompare(v1,v2,tree->sblk->keyLen));*/
			return(memcmp(v1,v2,tree->sblk->keyLen));
		case IDX_CHAR:
			return(charCompare(v1,v2));
		case IDX_INT8:
			return(int8Compare(v1,v2));
		case IDX_INT16:
			return(int16Compare(v1,v2));
		case IDX_INT32:
			return(int32Compare(v1,v2));
		case IDX_INT64:
			return(int64Compare(v1,v2));
		case IDX_UINT8:
			return(uint8Compare(v1,v2));
		case IDX_UINT16:
			return(uint16Compare(v1,v2));
		case IDX_UINT32:
			return(uint32Compare(v1,v2));
		case IDX_UINT64:
			return(uint64Compare(v1,v2));
		case IDX_REAL:
			return(realCompare(v1,v2));
		default:
			/*return(byteCompare(v1,v2,tree->sblk->keyLen));*/
			return(memcmp(v1,v2,tree->sblk->keyLen));
	}
}


static void swapNodes(tree, src, dst)
	avltree	*tree;
	register avl_nod	*src,
		*dst;
{
	int	tmpLeft,
		tmpRight,
		tmpParent,
		tmpHeight;
	avl_nod	*parent,
		*tmp;

	/*
	** Re-link the nodes
	*/
	tmpLeft = dst->left;
	tmpRight = dst->right;
	tmpParent = dst->parent;
	tmpHeight = dst->height;

	dst->left = src->left;
	dst->right = src->right;
	dst->parent = src->parent;
	dst->height = src->height;

	src->left = tmpLeft;
	src->right = tmpRight;
	src->parent = tmpParent;
	src->height = tmpHeight;

	/*
	** Check for wierd links left over from swapping neighbours
	*/
	if (dst->left == dst->nodeNum)
		dst->left = src->nodeNum;
	if (dst->right == dst->nodeNum)
		dst->right = src->nodeNum;
	if (dst->parent == dst->nodeNum)
		dst->parent = src->nodeNum;
	if (src->left == src->nodeNum)
		src->left = dst->nodeNum;
	if (src->right == src->nodeNum)
		src->right = dst->nodeNum;
	if (src->parent == src->nodeNum)
		src->parent = dst->nodeNum;
	


	/*
	** Relink the parents
	*/
	if (src->parent != dst->nodeNum)
	{
		parent = getTreeNode(tree, src->parent);
		if (parent)
		{
			if (parent->left == dst->nodeNum)
				parent->left = src->nodeNum;
			else
				parent->right = src->nodeNum;
		}
		else
		{
			tree->sblk->root = src->nodeNum;
		}
	}
	if (dst->parent != src->nodeNum)
	{
		parent = getTreeNode(tree,dst->parent);
		if (parent)
		{
			if (parent->left == src->nodeNum)
				parent->left = dst->nodeNum;
			else
				parent->right = dst->nodeNum;
		}
		else
		{
			tree->sblk->root = dst->nodeNum;
		}
	}

	/*
	** Relink the kids
	*/

	if (src->left != dst->nodeNum)
	{
		tmp = getTreeNode(tree,src->left);
		if (tmp)
			tmp->parent = src->nodeNum;
	}
	if (src->right != dst->nodeNum)
	{
		tmp = getTreeNode(tree,src->right);
		if (tmp)
			tmp->parent = src->nodeNum;
	}
	if (dst->left != dst->nodeNum)
	{
		tmp = getTreeNode(tree,dst->left);
		if (tmp)
			tmp->parent = dst->nodeNum;
	}
	if (dst->right != dst->nodeNum)
	{
		tmp = getTreeNode(tree,dst->right);
		if (tmp)
			tmp->parent = dst->nodeNum;
	}
}


/*************************************************************************
**************************************************************************
**                                                                      **
**                        INTERFACE ROUTINES                            **
**                                                                      **
**************************************************************************
*************************************************************************/


int avlCreate(path,mode,keyLen,keyType,flags,version)
	char	*path;
	int	mode,
		keyLen,
		keyType,
		flags,
		version;
{
	int	fd;
	avl_sbk	sblk;
	ptr_sbk	psblk;
	char	blank[1],
		*tmpPath;

	/*
	** Create the tree file
	*/
	fd = open(path,O_RDWR | O_CREAT, mode);
	if (fd < 0)
	{
		return(-1);
	}

	/*
	** Write the tree superblock
	*/
	bzero(&sblk, sizeof(avl_sbk));
	if (version == AVL_V1)
		sblk.magic = AVL_MAGIC_V1;
	if (version == AVL_V2)
		sblk.magic = AVL_MAGIC_V2;
	sblk.keyType = keyType;
	if (keyType == IDX_CHAR)
		keyLen++;
	sblk.keyLen = keyLen;
	sblk.flags = flags;
	sblk.nodeLen = DISK_NODE_SIZE(keyLen) + (8 -
                (DISK_NODE_SIZE(keyLen) % 8));

	if (write(fd,&sblk,sizeof(sblk)) < sizeof(sblk))
	{
		close(fd);
		unlink(path);
		return(-1);
	}
	if (sizeof(sblk) < SBLK_SIZE)
	{
		*blank = 0;
		lseek(fd,SBLK_SIZE-1,SEEK_SET);
		write(fd,blank,1);
	}
	close(fd);

	/*
	** Create the pointer file with a PSblk
	*/
	tmpPath = malloc(strlen(path) + 5);
	sprintf(tmpPath,"%s.ptr", path);
	fd = open(tmpPath,O_RDWR | O_CREAT, mode);
	free(tmpPath);
	if (fd < 0)
	{
		unlink(path);
		return(-1);
	}

	bzero(&psblk, sizeof(ptr_sbk));
	if (version == AVL_V1)
		psblk.magic = AVL_MAGIC_V1;
	if (version == AVL_V2)
		psblk.magic = AVL_MAGIC_V2;
	if (write(fd,&psblk,sizeof(psblk)) < sizeof(psblk))
	{
		close(fd);
		unlink(path);
		return(-1);
	}
	if (sizeof(psblk) < PSBLK_SIZE)
	{
		*blank = 0;
		lseek(fd,PSBLK_SIZE-1,SEEK_SET);
		write(fd,blank,1);
	}
	close(fd);
	return(0);
}



avltree *avlMemCreate(keyLen,keyType,flags)
	int	keyLen,
		keyType,
		flags;
{
	avltree	*new;
	int	nodeLen;


	/*
	** Create the base contents
	*/
	new = (avltree *)malloc(sizeof(avltree));
	bzero(new, sizeof(avltree));
	nodeLen = DISK_NODE_SIZE(keyLen) + (8 - (DISK_NODE_SIZE(keyLen) % 8));
	new->memLen = SBLK_SIZE;
	new->memTree = (char *)malloc(new->memLen);
	bzero(new->memTree, new->memLen);

	new->sblk = (avl_sbk *)new->memTree;
	new->sblk->magic = AVL_MAGIC_V2;
	new->sblk->keyType = keyType;
	if (keyType == IDX_CHAR)
		keyLen++;
	new->sblk->keyLen = keyLen;
	new->sblk->flags = flags;
	new->sblk->nodeLen = DISK_NODE_SIZE(keyLen) + (8 -
                (DISK_NODE_SIZE(keyLen) % 8));

	new->treeMapLen = new->memLen;
	new->treeMap = new->memTree;
	new->version = 1;
	return(new);
}


avltree *avlOpen(path)
	char	*path;
{
	int	fd;
	avltree	*new;
	struct	stat sbuf;
	char	*tmpPath;


	/*
	** Open and map the tree file
	*/
	fd = open(path,O_RDWR, 0);
	if (fd < 0)
	{
		return(NULL);
	}

	if (fstat(fd, &sbuf) < 0)
	{
		return(NULL);
	}
	new = (avltree *)malloc(sizeof(avltree));
	bzero(new,sizeof(avltree));
	new->treeFd = fd;
	strncpy(new->path, path, AVL_PATH_LEN);
	new->treeMapLen = sbuf.st_size;
	new->treeMap = mmap(NULL, new->treeMapLen, PROT_READ|PROT_WRITE,
		MAP_SHARED, new->treeFd, (off_t)0);
	if (new->treeMap == (caddr_t)-1)
	{
		close(new->treeFd);
		free(new);
		return(NULL);
	}
	new->sblk = (avl_sbk *)new->treeMap;
	new->sblk->fileLen = new->treeMapLen;

	
	if (new->sblk->magic != AVL_MAGIC_V1 &&
		new->sblk->magic != AVL_MAGIC_V2)
	{
		/* Invalid file magic */
		close(new->treeFd);
		free(new);
		return(NULL);
	}
	if (new->sblk->magic == AVL_MAGIC_V1)
	{
		new->version = 1;
		new->ptrFd = -1;
		return(new);
	}

	/*
	** Must be a version 2 formatted tree so map the ptr file as well
	*/
	new->version = 2;
	tmpPath = malloc(strlen(path) + 5);
	sprintf(tmpPath, "%s.ptr", path);
	fd = open(tmpPath,O_RDWR, 0);
	free(tmpPath);
	if (fd < 0)
	{
		perror("ptr open");
		close(new->treeFd);
		free(new);
		return(NULL);
	}
	new->ptrFd = fd;
	if (fstat(fd, &sbuf) < 0)
	{
		perror("ptr stat");
		close(new->treeFd);
		close(new->ptrFd);
		free(new);
		return(NULL);
	}
	new->ptrMapLen = sbuf.st_size;
	if (new->ptrMapLen == -1)
	{
		perror("ptr stat size");
		close(new->treeFd);
		close(new->ptrFd);
		free(new);
		return(NULL);
	}
	if (new->ptrMapLen == 0)
	{
		new->ptrMap = NULL;
		new->psblk = NULL;
		return(new);
	}
	new->ptrMap = mmap(NULL, new->ptrMapLen, PROT_READ|PROT_WRITE,
		MAP_SHARED, new->ptrFd, (off_t)0);
	if (new->ptrMap == (caddr_t)-1)
	{
		perror("ptr mmap");
		close(new->treeFd);
		close(new->ptrFd);
		free(new);
		return(NULL);
	}
	new->psblk = (ptr_sbk *)new->ptrMap;
	return(new);
}


void avlClose(tree)
	avltree	*tree;
{
	if (tree == NULL)
	{
		return;
	}
	if (tree->memTree == NULL)
	{
		if (tree->treeMapLen)
			munmap(tree->treeMap, tree->treeMapLen);
		close(tree->treeFd);
		if (tree->version == 2)
		{
			if (tree->ptrMapLen)
				munmap(tree->ptrMap, tree->ptrMapLen);
			close(tree->ptrFd);
		}
	}
	else
	{
		free(tree->memTree);
		tree->memTree = NULL;
	}
	free(tree);
}



void avlSync(tree)
	avltree	*tree;
{
	if (checkTreeRemap(tree) < 0)
	{
		return;
	}

	if (tree == NULL)
		return;
	if (tree->treeMapLen)
		MSYNC(tree->treeMap, tree->treeMapLen, 0);
	if (tree->ptrMapLen)
		MSYNC(tree->ptrMap, tree->ptrMapLen, 0);
}



static avl_nod *_avlLookup(tree, key, flags)
	avltree	*tree;
	char	*key;
	int	flags;
{
	REG 	int	res;
	REG	avl_nod	*cur;


	if (AVL_DEBUG)
	{
		printf("AVL Lookup of ");
		avlPrintElement(key,tree->sblk->keyType,tree->sblk->keyLen);
		printf(" in tree %s\n", tree->memTree ? "MEMTREE" :
			rindex(tree->path,'/'));
	}

	tree->curNode = 0;
	cur = getTreeNode(tree,tree->sblk->root);
	while(cur)
	{
		tree->curNode = cur->nodeNum;
		res = compareValues(key, nodeKey(cur), tree);
		if (AVL_DEBUG > 1)
		{
			printf("\tCompare node %d ",cur->nodeNum);
			avlPrintElement(nodeKey(cur),tree->sblk->keyType,
				tree->sblk->keyLen);
			printf(".  Result = %d\n",res);
		}
		if (res > 0)
		{
			if (cur->right == 0)
			{
				if (flags & IDX_EXACT)
				{
					return(NULL);
				}
				else
				{
					cur->key = nodeKey(cur);
					return(cur);
				}
			}
			cur = getTreeNode(tree,cur->right);
			continue;
		}
		if (res < 0)
		{
			if (cur->left == 0)
			{
				if (flags & IDX_EXACT)
				{
					return(NULL);
				}
				else
				{
					cur->key = nodeKey(cur);
					return(cur);
				}
			}
			cur = getTreeNode(tree,cur->left);
			continue;
		}
		cur->key = nodeKey(cur);
		return(cur);
	}
	return(NULL);
}



avl_nod *avlLookup(tree, key, flags)
	avltree	*tree;
	char	*key;
	int	flags;
{
	if (checkTreeRemap(tree) < 0)
	{
		return(NULL);
	}
	return(_avlLookup(tree,key,flags));
}



void avlSetCursor(tree, cursor)
	avltree	*tree;
	avl_cur	*cursor;
{
	cursor->curNode = tree->curNode;
	cursor->curDup = 0;
	cursor->nextDup = 0;
}



avl_nod *avlGetFirst(tree)
	avltree	*tree;
{
	avl_nod	*cur;

	if (checkTreeRemap(tree) < 0)
	{
		return(NULL);
	}

	cur = getTreeNode(tree,tree->sblk->root);
	tree->curNode = 0;
	if (!cur)
		return(NULL);
	while(cur->left)
	{
		cur = getTreeNode(tree,cur->left);
	}
	tree->curNode = cur->nodeNum;
	return(cur);
}


avl_nod *avlGetLast(tree)
	avltree	*tree;
{
	avl_nod	*cur;

	if (checkTreeRemap(tree) < 0)
	{
		return(NULL);
	}

	cur = getTreeNode(tree,tree->sblk->root);
	tree->curNode = 0;
	if (!cur)
		return(NULL);
	while(cur->right)
	{
		cur = getTreeNode(tree,cur->right);
	}
	tree->curNode = cur->nodeNum;
	return(cur);
}



avl_nod *avlGetNext(tree,cursor)
	avltree	*tree;
	avl_cur	*cursor;
{
	avl_nod	*cur,
		*prev;
	int	prevNode;


	if (checkTreeRemap(tree) < 0)
	{
		return(NULL);
	}

	/*
	** Are we somewhere down a dup chain?
	*/
	if (cursor->curDup != 0)
	{
		if (AVL_DEBUG > 2)
		{
			printf("AVL GetNext from node dup %u in tree %s\n",
				cursor->curDup, tree->memTree ? "MEMTREE" :
				rindex(tree->path,'/'));
		}
		cur = getTreeNode(tree,cursor->curDup);
#ifdef NOTDEF
		if (cur->dup != cursor->nextDup)
		{
			if (AVL_DEBUG > 2)
			{
				printf("\tDup chain edited, using node %u\n", 
					cur->nodeNum);
			}
			/* Dup chain has been edited. */
			return(cur);
			
		}
#endif
		if (cur->dup)
		{
			cur = getTreeNode(tree,cur->dup);
			cursor->curDup = cur->nodeNum;
			cursor->nextDup = cur->dup;
			if (AVL_DEBUG > 2)
			{
				printf("\tUsing next cursor dup %u\n", 
					cur->nodeNum);
			}
			cur->key = nodeKey(cur);
			return(cur);
		}
	}
	cur = getTreeNode(tree,cursor->curNode);
	if (cur->dup != 0 && cursor->curDup == 0)
	{
		cur = getTreeNode(tree,cur->dup);
		cursor->curDup = cur->nodeNum;
		cursor->nextDup = cur->dup;
		if (AVL_DEBUG > 2)
		{
			printf("\tUsing node dup %u\n", cur->nodeNum);
		}
		cur->key = nodeKey(cur);
		return(cur);
	}


	/*
	** No dups to return so move onto the next node
	*/
	if (cur->right)
	{
		cur = getTreeNode(tree,cur->right);
		while(cur->left)
		{
			prev = cur;
			cur = getTreeNode(tree,cur->left);
		}
		cursor->curNode = cur->nodeNum;
		cursor->curDup = 0;
		cursor->nextDup = 0;
		if (AVL_DEBUG > 2)
		{
			printf("\tUsing right child %u\n", cur->nodeNum);
		}
		cur->key = nodeKey(cur);
		return(cur);
	}

	while (cur->parent)
	{
		prevNode = cur->nodeNum;
		cur = getTreeNode(tree,cur->parent);
		if (cur->left == prevNode)
		{
			cursor->curNode = cur->nodeNum;
			cursor->curDup = 0;
			cursor->nextDup = 0;
			if (AVL_DEBUG > 2)
			{
				printf("\tTrace back up the tree gave %u\n", 
					cur->nodeNum);
			}
			cur->key = nodeKey(cur);
			return(cur);
		}
	}
	cursor->curNode = cursor->curDup = cursor->nextDup = 0;
	return(NULL);
}



avl_nod *avlGetPrev(tree,cursor)
	avltree	*tree;
	avl_cur	*cursor;
{
	avl_nod	*cur;
	int	prevNode;

	/*
	** Are we somewhere down a dup chain?
	*/

	if (checkTreeRemap(tree) < 0)
	{
		return(NULL);
	}

	if (cursor->curDup != 0)
	{
		cur = getTreeNode(tree,cursor->curDup);
		if (cur->dup)
		{
			cur = getTreeNode(tree,cur->dup);
			cursor->curDup = cur->nodeNum;
			cur->key = nodeKey(cur);
			return(cur);
		}
	}
	cur = getTreeNode(tree,cursor->curNode);
	if (cur->dup != 0 && cursor->curDup == 0)
	{
		cursor->curDup = cur->dup;
		cur = getTreeNode(tree,cur->dup);
		cur->key = nodeKey(cur);
		return(cur);
	}

	cur = getTreeNode(tree,cursor->curNode);
	if (cur->left)
	{
		cur = getTreeNode(tree,cur->left);
		while(cur->right)
		{
			cur = getTreeNode(tree,cur->right);
		}
		cursor->curNode = cur->nodeNum;
		cursor->curDup = 0;
		cur->key = nodeKey(cur);
		return(cur);
	}

	while (cur->parent)
	{
		prevNode = cur->nodeNum;
		cur = getTreeNode(tree,cur->parent);
		if (cur->right == prevNode)
		{
			cursor->curNode = cur->nodeNum;
			cursor->curDup = 0;
			cur->key = nodeKey(cur);
			return(cur);
		}
	}
	cursor->curNode = 0;
	cursor->curDup = 0;
	return(NULL);
}


int avlInsert(tree, key, data)
	avltree	*tree;
	char	*key;
	off_t	data;
{
	avl_nod	*cur,
		*new,
		*tmp;
	int	res,
		debug,
		oldHeight;




	/*
	** Find the insertion point
	*/
	if (checkTreeRemap(tree) < 0)
	{
		return(-1);
	}

	new = getFreeNode(tree);
	if (new == NULL)
	{
		return(IDX_UNKNOWN);
	}


	/*
	** Find the closest node for this value. Turn off debugging so
	** we don't see a lookup in the middle of an insert
	*/
	debug = AVL_DEBUG;
	AVL_DEBUG = 0;
	cur = avlLookup(tree, key, IDX_CLOSEST);
	AVL_DEBUG = debug;


	if (!cur)
	{
		/*
		** Tree must be empty
		*/
		new->left = new->right = new->parent = new->height = 0;
		copyValue(key, nodeKey(new), tree);
		new->data = data;
		tree->sblk->root = new->nodeNum;
		tree->sblk->numEntries = 1;
		tree->sblk->numKeys = 1;
		if (tree->version > 1)
		{
			if (addPtrEntry(tree, data, new->nodeNum) < 0)
			{
				return(-1);
			}
		}
		return(IDX_OK);
	}

	/*
	** Add the node
	*/
	res = compareValues(key,nodeKey(cur),tree);

	if (AVL_DEBUG)
	{
		printf("AVL Insert of value ");
		avlPrintElement(key,tree->sblk->keyType,tree->sblk->keyLen);
		printf(", data %u at %s node %u in tree %s\n", data,
			res==0?"dup":"tree", new->nodeNum, 
			tree->memTree ? "MEMTREE" : rindex(tree->path,'/'));
	}

	if (res == 0)
	{
		/*
		** Duplicate Value
		*/
		new->dup = cur->dup;
		cur->dup = new->nodeNum;
		new->height = -1;
		new->data = data;
		new->parent = cur->nodeNum;  /* tree node */

		/*
		** The parent of each dup entry is the node before
		** it in the dup chain.  Reset the parent of the head of
		** the original dup chain if there was one.
		*/
		if(new->dup)
		{
			tmp = getTreeNode(tree,new->dup);
			tmp->parent = new->nodeNum;
		}

		copyValue(key, nodeKey(new), tree);
		tree->sblk->numEntries++;
		if (tree->version > 1 &&
			addPtrEntry(tree, data, new->nodeNum) < 0)
		{
			return(-1);
		}
		if (TEST_TREE)
		{
			avlTestTree(tree);
		}
		return(IDX_OK);
	}

	/*
	** It's not a dup
	*/
	if (res < 0)
		cur->left = new->nodeNum;
	else
		cur->right = new->nodeNum;
	new->left = new->right = new->height = 0;
	copyValue(key, nodeKey(new), tree);
	new->data = data;
	new->parent = cur->nodeNum;
	tree->sblk->numEntries++;
	tree->sblk->numKeys++;

	/*
	** Update height values as required if this is a new layer
	** (i.e. this is the first node placed under our parent)
	*/
	if (cur->left == 0 || cur->right == 0)
	{
		while(cur)
		{
			oldHeight = cur->height;
			setNodeHeight(tree,cur);
			if (cur->height == oldHeight)
				break;
			balanceTree(tree,cur);
			cur = getTreeNode(tree,cur->parent);
		}
	}

	/*
	** Update the pointer file if this is a version 2 or better AVL
	*/
	if (tree->version == 1)
	{
		if (TEST_TREE)
		{
			avlTestTree(tree);
		}
		return(IDX_OK);
	}
	if (addPtrEntry(tree, data, new->nodeNum) < 0)
	{
		return(-1);
	}

	if (TEST_TREE)
	{
		avlTestTree(tree);
	}
	return(IDX_OK);
}




int avlDelete(tree, key, data)
	avltree	*tree;
	char	*key;
	off_t	data;
{
	avl_nod	*cur,
		*parent = NULL,
		*next = NULL,
		*prev,
		*child,
		*tmp;
	avl_cur	cursor;
	int	nodeNum,
		debug,
		tmpInt,
		done = 0,
		oldHeight;
	char	*cp;
	off_t	offset;

	if (AVL_DEBUG)
	{
		printf("AVL Delete of ");
		avlPrintElement(key,tree->sblk->keyType,tree->sblk->keyLen);
		printf(" : %u tree %s\n", data, tree->memTree ? "MEMTREE" :
			rindex(tree->path,'/'));
	}

	/*
	** Find the node that matches both the key and the data.  If
	** we have a Pointer file then we can sidestep the tree lookup
	*/
	if (checkTreeRemap(tree) < 0)
	{
		return(-1);
	}

	if (tree->version > 1 && tree->ptrMapLen > 0)
	{
		/*
		** Lookup the pointer file for this data value
		** and load the corresponding tree node
		*/
		nodeNum = getPtrEntry(tree, data);
		if (nodeNum < 0)
		{
			return(-1);
		}
		cur = getTreeNode(tree, nodeNum);
                tree->curNode = nodeNum;
		avlSetCursor(tree,&cursor);
		if (delPtrEntry(tree, data) < 0)
		{
			return(-1);
		}

		if (AVL_DEBUG > 1)
		{
			printf("\tPointer lookup gave node %d\n", nodeNum);
		}
		/*
		** If we are using a pointer file then we must ensure
		** that node data is not moved between nodes as the pointer
		** maps data values directly to tree nodes.  Move the node
		** instead.
		*/
		if (cur->height != -1 && cur->dup)
		{
			if (AVL_DEBUG > 1)
			{
				printf("\tNode %d is a tree node with dups %d.",
					nodeNum, cur->dup);
				if (cur->parent == 0)
					printf(" Node is the ROOT\n");
				else
					printf(" Parent is %d\n", cur->parent);
			}
	
			/* 
			** It's a tree node with a dup chain.  Slot the
			** first entry on the dup chain into the tree in
			** place of this node.  Repoint the pointers in
			** the parent and child nodes.
			*/
			tmp = getTreeNode(tree,cur->dup);
			parent = getTreeNode(tree,cur->parent);
			tmp->left = cur->left;
			tmp->right = cur->right;
			tmp->parent = cur->parent;
			tmp->height = cur->height;
			copyValue(nodeKey(cur),nodeKey(tmp),tree);

			if (parent)
			{
				if (parent->left == cur->nodeNum)
					parent->left = tmp->nodeNum;
				else
					parent->right = tmp->nodeNum;
			}
			else
			{
				/* This is now the tree root */
				tree->sblk->root = tmp->nodeNum;
			}

			if (cur->left)
			{
				child = getTreeNode(tree,cur->left);
				child->parent = tmp->nodeNum;
			}
			if (cur->right)
			{
				child = getTreeNode(tree,cur->right);
				child->parent = tmp->nodeNum;
			}
			dropNode(tree,cur);
			tree->sblk->numEntries--;
			if (TEST_TREE)
			{
				avlTestTree(tree);
			}
			return(IDX_OK);
		}
		if (cur->height == -1)
		{
			if (AVL_DEBUG > 1)
			{
				printf("\tNode %d is part of a dup chain. "
					"Next dup %d\n",nodeNum, cur->dup);
			}
			/* 
			** It's part of a dup chain.  Remove it from
			** the chain
			*/
			parent = getTreeNode(tree,cur->parent);
			parent->dup = cur->dup;
			if (cur->dup)
			{
				tmp = getTreeNode(tree,cur->dup);
				tmp->parent = parent->nodeNum;
			}
			dropNode(tree,cur);
			tree->sblk->numEntries--;
			if (TEST_TREE)
			{
				avlTestTree(tree);
			}
			return(IDX_OK);
		}
	}
	else
	{

		/*
		** No Pointer file so do it the hard way.  Perform a lookup
		** for the value in the tree but turn off debugging so we
		** don't see a lookup in the middle of a delete debug output.
		*/
		debug = AVL_DEBUG;
		AVL_DEBUG = 0;
		cur = avlLookup(tree,key,IDX_EXACT);
		AVL_DEBUG = debug;
		if (!cur)
			return(IDX_NOT_FOUND);
		avlSetCursor(tree,&cursor);

		if (cur->dup)
		{
			if (cur->data == data)
			{
				tmp = getTreeNode(tree,cur->dup);
				cur->dup = tmp->dup;
				cur->data = tmp->data;
				dropNode(tree,tmp);
				if (cur->dup)
				{
					tmp = getTreeNode(tree,cur->dup);
					tmp->parent = cur->nodeNum;
				}
				tree->sblk->numEntries--;
				if (TEST_TREE)
				{
					avlTestTree(tree);
				}
				return(IDX_OK);
			}
			prev = cur;
			cur = getTreeNode(tree,cur->dup);
			while(cur)
			{
				if (cur->data == data)
				{
					prev->dup = cur->dup;
					if (cur->dup)
					{
						tmp = getTreeNode(tree,cur->dup);
						tmp->parent = prev->nodeNum;
					}
					tree->sblk->numEntries--;
					dropNode(tree,cur);
					return(IDX_OK);
				}
				prev = cur;
				cur = getTreeNode(tree,cur->dup);
			}
			if (TEST_TREE)
			{
				avlTestTree(tree);
			}
			return(IDX_NOT_FOUND);
		}
	}

	/*
	** OK, so I'm a whimp!  Let's take the ultra easy option on
	** this.  What we want is for this node to be a leaf node with
	** no kids.  Lets just keep swapping it with it's next greater
	** or lesser node until it becomes a leaf.  We want to ensure
	** that the swap direction is down the tree so check our kids.
	*/

	cursor.curNode = cur->nodeNum;
	cursor.curDup = 0;

	if (cur->right)
	{
		while(cur->left || cur->right)
		{
			next = avlGetNext(tree,&cursor);
			if (!next)
				break;
			swapNodes(tree, cur, next);
			avlSetCursor(tree,&cursor);
		}
		if (!next)
		{
			/*
			** Hmmm, can't get away from the child link.
			** We know that there's only 1 child though so
			** we can just collapse this branch
			*/
			next = getTreeNode(tree,cur->left);
			next->parent = cur->parent;
			parent = getTreeNode(tree,cur->parent);
			if (parent->left == cur->nodeNum)
				parent->left = next->nodeNum;
			else
				parent->right = next->nodeNum;
			dropNode(tree,cur);
			tree->sblk->numEntries--;
			tree->sblk->numKeys--;
			done = 1;
		}
	}
	else if (cur->left)
	{
		while(cur->left || cur->right)
		{
			next = avlGetPrev(tree,&cursor);
			if (!next)
				break;
			swapNodes(tree, cur, next);
			avlSetCursor(tree,&cursor);
		}
		if (!next)
		{
			/*
			** As above
			*/
			next = getTreeNode(tree,cur->right);
			next->parent = cur->parent;
			parent = getTreeNode(tree,cur->parent);
			if (parent->left == cur->nodeNum)
				parent->left = next->nodeNum;
			else
				parent->right = next->nodeNum;
			dropNode(tree,cur);
			tree->sblk->numEntries--;
			tree->sblk->numKeys--;
			done = 1;
		}
	}

	/*
	** By this time it must be a leaf node.
	*/
	if (!done)
	{
		parent = getTreeNode(tree,cur->parent);
		if (!parent)
		{
			/*
			** It's the last node in the tree
			*/
			dropNode(tree,cur);
			tree->sblk->root = 0;
			tree->sblk->numEntries = tree->sblk->numKeys = 0;
			return(IDX_OK);
		}
		if (parent->left == cur->nodeNum)
		{
			parent->left = 0;
			if (parent->right == 0)
			{
				parent->height = 0;
				parent = getTreeNode(tree, parent->parent);
			}
			dropNode(tree,cur);
			tree->sblk->numEntries--;
			tree->sblk->numKeys--;
		}
		else if (parent->right == cur->nodeNum)
		{
			parent->right = 0;
			if (parent->left == 0)
			{
				parent->height = 0;
				parent = getTreeNode(tree, parent->parent);
			}
			dropNode(tree,cur);
			tree->sblk->numEntries--;
			tree->sblk->numKeys--;
		}
	}

	/*
	** Rebalance as required
	*/
	while(parent)
	{
		oldHeight = parent->height;
		setNodeHeight(tree,parent);
		if (parent->height == oldHeight)
			break;
		balanceTree(tree,parent);
		parent = getTreeNode(tree,parent->parent);
	}
	if (TEST_TREE)
	{
		avlTestTree(tree);
	}
	return(IDX_OK);
}



/**************************************************************************
** Debugging Code
*/

static void printNode(tree, cur, depth, vFlag)
	avltree	*tree;
	avl_nod	*cur;
	int	depth,
		vFlag;
{
	if (vFlag)
	{
		showIndent(depth);
		printf("N=%d, P=%d, L=%d, R=%d H=%d D=%d\n",cur->nodeNum, 
			cur->parent, cur->left, cur->right, cur->height,
			cur->dup);
		showIndent(depth);
	}
	avlPrintElement(nodeKey(cur), tree->sblk->keyType,tree->sblk->keyLen);
	printf("data = %d\n", (int)cur->data);
}



static void dumpSubTree(tree, node, depth, vFlag)
	avltree	*tree;
	int	node,
		depth,
		vFlag;
{
	avl_nod	*cur;

	if (node == 0)
		return;

	cur = getTreeNode(tree, node);
	dumpSubTree(tree, cur->left, depth+1, vFlag);
	printNode(tree, cur, depth, vFlag);
	dumpSubTree(tree, cur->right, depth+1, vFlag);
}



void avlDumpTree(tree, verbose)
	avltree	*tree;
	int	verbose;
{
	dumpSubTree(tree,tree->sblk->root, 0,verbose);
}



int avlTestTree(tree)
	avltree	*tree;
{
	avl_nod	*cur,
		*prev = NULL;
	avl_cur	cursor;
	int	count,
		prev1 = 0, prev2 = 0, prev3 = 0, prev4 = 0,
		prev5 = 0, prev6 = 0, prev7 = 0, prev8 = 0,
		prev9 = 0, prev10 = 0, prev11 = 0, prev12 = 0,
		prev13 = 0, prev14 = 0, prev15 = 0, prev16 = 0,
		prev17 = 0, prev18 = 0, prev19 = 0, prev20 = 0,
		prev21 = 0, prev22 = 0, prev23 = 0, prev24 = 0;

	cur = avlGetFirst(tree);
	if (!cur)
		return(0);
	count = 1;
	avlSetCursor(tree,&cursor);
	prev = cur;
	cur = avlGetNext(tree,&cursor);
	while(cur)
	{
		if (cur->nodeNum == prev1 || cur->nodeNum == prev2 ||
			cur->nodeNum == prev3 || cur->nodeNum == prev4 ||
			cur->nodeNum == prev5 || cur->nodeNum == prev6 ||
			cur->nodeNum == prev7 || cur->nodeNum == prev8 ||
			cur->nodeNum == prev9 || cur->nodeNum == prev10 ||
			cur->nodeNum == prev11 || cur->nodeNum == prev12 ||
			cur->nodeNum == prev13 || cur->nodeNum == prev14 ||
			cur->nodeNum == prev15 || cur->nodeNum == prev16 ||
			cur->nodeNum == prev17 || cur->nodeNum == prev18 ||
			cur->nodeNum == prev19 || cur->nodeNum == prev20 ||
			cur->nodeNum == prev21 || cur->nodeNum == prev13 ||
			cur->nodeNum == prev23 || cur->nodeNum == prev14 ||
			cur->nodeNum == cur->parent || 
			count > tree->sblk->numEntries)
		{
			abort();
		}
		count++;
		if (compareValues(nodeKey(prev), nodeKey(cur), tree) > 0)
		{
			return(-1);
		}
		prev = cur;
		prev24 = prev23;
		prev23 = prev22;
		prev22 = prev21;
		prev21 = prev20;
		prev20 = prev19;
		prev19 = prev18;
		prev18 = prev17;
		prev17 = prev16;
		prev16 = prev15;
		prev15 = prev14;
		prev14 = prev13;
		prev13 = prev12;
		prev12 = prev11;
		prev11 = prev10;
		prev10 = prev9;
		prev9 = prev8;
		prev8 = prev7;
		prev7 = prev6;
		prev6 = prev5;
		prev5 = prev4;
		prev4 = prev3;
		prev3 = prev2;
		prev2 = prev1;
		prev1 = cur->nodeNum;
		cur = avlGetNext(tree,&cursor);
	}
	return(count);
}
